import time

import flask

from ..client import IntrospectionClient


class FlaskIntrospection:
    def __init__(self, client: IntrospectionClient) -> None:
        self._client = client

    def register(self, app: flask.Flask) -> None:
        app.before_request(self.before_request)
        app.after_request(self.after_request)

    def before_request(self) -> None:
        flask.g.start_time = time.time()

    def after_request(self, response: flask.Response) -> flask.Response:
        self._client.send(
            self._client.make_checkpoint(
                "flask",
                {
                    "method": flask.request.method,
                    "url": flask.request.url,
                    "headers": dict(flask.request.headers),
                    "body": flask.request.get_data(as_text=True),
                },
                {
                    "status_code": response.status_code,
                    "headers": dict(response.headers),
                    "body": response.get_data(as_text=True),
                },
                time.time() - flask.g.start_time,
            )
        )
        return response



import json
import logging
import time
from functools import wraps

import requests

logger = logging.getLogger(__name__)


class IntrospectionClient:
    def __init__(self, session: requests.Session, api_url: str, name: str) -> None:
        self.api_url = api_url
        self._session = session
        self._name = name

    def send(self, data: dict) -> None:
        try:
            response = self._session.post(self.api_url, json=data)
            if response.status_code != 200:
                logger.error("Failed to send checkpoint: %r", response.text)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error sending checkpoint")

    def make_checkpoint(
        self, location: str, input_data: dict, output_data: dict, duration: float
    ) -> dict:
        return {
            "location": f"{self._name}/{location}",
            "timestamp": time.time(),
            "input": json.dumps(input_data),
            "output": json.dumps(output_data),
            "metadata": {"duration": duration},
        }


class IntrospectionDecorator:
    instance = None

    def __init__(self, client: IntrospectionClient) -> None:
        self._client = client

    def register(self) -> None:
        self.__class__.instance = self

    def unregister(self) -> None:
        self.__class__.instance = None

    def wrapper(self, func, *args, **kwargs):
        start_time = time.time()
        try:
            result = func(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            self._client.send(
                self._client.make_checkpoint(
                    func.__name__,
                    {"args": args, "kwargs": kwargs},
                    {"error": str(exc)},
                    time.time() - start_time,
                )
            )
            raise
        self._client.send(
            self._client.make_checkpoint(
                func.__name__,
                {"args": args, "kwargs": kwargs},
                result,
                time.time() - start_time,
            )
        )
        return result


def introspect(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if IntrospectionDecorator.instance:
            return IntrospectionDecorator.instance.wrapper(func, *args, **kwargs)
        return func(*args, **kwargs)

    return wrapper



import logging
import unittest
from unittest import mock

import requests

from .client import IntrospectionClient, IntrospectionDecorator, introspect, logger


class TestIntrospectionClient(unittest.TestCase):
    def setUp(self) -> None:
        logger.setLevel(logging.ERROR)
        logger.addHandler(logging.StreamHandler())

    def test_send_successful(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.return_value.status_code = 200
        client = IntrospectionClient(mock_session, "http://example.com", "test_client")

        # Act
        client.send({"key": "value"})

        # Assert
        mock_session.post.assert_called_once_with("http://example.com", json={"key": "value"})

    def test_send_failure(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.return_value.status_code = 500
        client = IntrospectionClient(mock_session, "http://example.com", "test_client")

        # Act
        with self.assertLogs(logger=logger, level="DEBUG") as log:
            client.send({"key": "value"})

        # Assert
        self.assertIn("ERROR:i8t.client:Failed to send checkpoint", log.output[0])

    def test_send_exception(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.side_effect = Exception("Test exception")
        client = IntrospectionClient(mock_session, "http://example.com", "test_client")

        # Act
        with self.assertLogs(logger=logger, level="ERROR") as log:
            client.send({"key": "value"})

        # Assert
        self.assertIn("ERROR:i8t.client:Error sending checkpoint", log.output[0])

    @mock.patch("time.time", return_value=1234567890)
    def test_make_checkpoint(self, _):
        # Arrange
        client = IntrospectionClient(None, "http://example.com", "test_client")

        # Act
        checkpoint = client.make_checkpoint("location1", {"input": "data"}, {"output": "data"}, 2.5)

        # Assert
        expected_checkpoint = {
            "location": "test_client/location1",
            "timestamp": 1234567890,
            "input": '{"input": "data"}',
            "output": '{"output": "data"}',
            "metadata": {"duration": 2.5},
        }
        self.assertEqual(checkpoint, expected_checkpoint)


class TestIntrospectionDecorator(unittest.TestCase):
    @mock.patch("time.time", side_effect=[1000, 2000])
    def test_wrapper_success(self, _):
        # Arrange
        mock_client = mock.Mock(IntrospectionClient)
        decorator = IntrospectionDecorator(mock_client)
        decorator.register()

        @introspect
        def test_func(first, second):
            return first + second

        # Act
        result = test_func(1, 2)

        # Assert
        self.assertEqual(result, 3)
        mock_client.send.assert_called_once()

        # Arrange
        decorator.unregister()

        # Act
        result = test_func(1, 2)

        # Assert
        self.assertEqual(result, 3)
        mock_client.send.assert_called_once()

    @mock.patch("time.time", side_effect=[1000, 2000])
    def test_wrapper_exception(self, _):
        # Arrange
        mock_client = mock.Mock(IntrospectionClient)
        decorator = IntrospectionDecorator(mock_client)
        decorator.register()

        @introspect
        def test_func(first, second):
            raise ValueError("Test exception")

        # Set up the return value for make_checkpoint mock
        mock_client.make_checkpoint.return_value = {
            "location": "test_func",
            "input": {"args": (1, 2), "kwargs": {}},
            "output": {"error": "Test exception"},
            "metadata": {"duration": 1000},
        }

        # Act & Assert
        with self.assertRaises(ValueError):
            test_func(1, 2)

        mock_client.send.assert_called_once()
        send_call_args = mock_client.send.call_args[0][0]
        self.assertIn("error", send_call_args["output"])



import time

import requests

from ..client import IntrospectionClient


class RequestsIntrospection:
    def __init__(self, client: IntrospectionClient) -> None:
        self._client = client
        self._original_request = requests.Session.request
        self._session = requests.Session()
        self._session_request = self._session.request

    def register(self) -> None:
        requests.Session.request = self._instrumented_request  # type: ignore[assignment]

    def unregister(self) -> None:
        requests.Session.request = self._original_request  # type: ignore[assignment]

    def _instrumented_request(self, method, url, **kwargs):
        if url == self._client.api_url:
            return self._session_request(method, url, **kwargs)
        start_time = time.time()
        try:
            response = self._session_request(method, url, **kwargs)
        except Exception as exc:
            self._client.send(
                self._client.make_checkpoint(
                    "requests",
                    {
                        "method": method,
                        "url": url,
                        "headers": dict(kwargs.get("headers", {})),
                        "body": kwargs.get("data", None),
                    },
                    {"error": str(exc)},
                    time.time() - start_time,
                )
            )
            raise
        self._client.send(
            self._client.make_checkpoint(
                "requests",
                {
                    "method": method,
                    "url": url,
                    "headers": dict(kwargs.get("headers", {})),
                    "body": kwargs.get("data", None),
                },
                {
                    "status_code": response.status_code,
                    "headers": dict(response.headers),
                    "body": response.text,
                },
                time.time() - start_time,
            )
        )
        return response
